from typing import Any, ClassVar, Literal, TextIO

from .json_context import ContextValues, JsonContext
from .object_comparer import ObjectComparer
from .string_file_wrapper import StringFileWrapper

JSONReturnType = dict[str, Any] | list[Any] | str | float | int | bool | None


class JSONParser:
    # Constants
    STRING_DELIMITERS: ClassVar[list[str]] = ['"', "'", "“", "”"]
    NUMBER_CHARS: ClassVar[set[str]] = set("0123456789-.eE/,")

    def __init__(
        self,
        json_str: str | StringFileWrapper,
        json_fd: TextIO | None,
        logging: bool | None,
        json_fd_chunk_length: int = 0,
        stream_stable: bool = False,
    ) -> None:
        # The string to parse
        self.json_str: str | StringFileWrapper = json_str
        # Alternatively, the file description with a json file in it
        if json_fd:
            # This is a trick we do to treat the file wrapper as an array
            self.json_str = StringFileWrapper(json_fd, json_fd_chunk_length)
        # Index is our iterator that will keep track of which character we are looking at right now
        self.index: int = 0
        # This is used in the object member parsing to manage the special cases of missing quotes in key or value
        self.context = JsonContext()
        # Use this to log the activity, but only if logging is active

        # This is a trick but a beautiful one. We call self.log in the code over and over even if it's not needed.
        # We could add a guard in the code for each call but that would make this code unreadable, so here's this neat trick
        # Replace self.log with a noop
        self.logging = logging
        if logging:
            self.logger: list[dict[str, str]] = []
            self.log = self._log
        else:
            # No-op
            self.log = lambda *args, **kwargs: None
        # When the json to be repaired is the accumulation of streaming json at a certain moment.
        # e.g. json obtained from llm response.
        # If this parameter to True will keep the repair results stable. For example:
        #   case 1:  '{"key": "val\\' => '{"key": "val"}'
        #   case 2:  '{"key": "val\\n' => '{"key": "val\\n"}'
        #   case 3:  '{"key": "val\\n123,`key2:value2' => '{"key": "val\\n123,`key2:value2"}'
        #   case 4:  '{"key": "val\\n123,`key2:value2`"}' => '{"key": "val\\n123,`key2:value2`"}'
        self.stream_stable = stream_stable

    def parse(
        self,
    ) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
        json = self.parse_json()
        if self.index < len(self.json_str):
            self.log(
                "The parser returned early, checking if there's more json elements",
            )
            json = [json]
            while self.index < len(self.json_str):
                j = self.parse_json()
                if j != "":
                    if ObjectComparer.is_same_object(json[-1], j):
                        # replace the last entry with the new one since the new one seems an update
                        json.pop()
                    json.append(j)
            # If nothing extra was found, don't return an array
            if len(json) == 1:
                self.log(
                    "There were no more elements, returning the element without the array",
                )
                json = json[0]
        if self.logging:
            return json, self.logger
        else:
            return json

    def parse_json(
        self,
    ) -> JSONReturnType:
        while True:
            char = self.get_char_at()
            # False means that we are at the end of the string provided
            if char is False:
                return ""
            # <object> starts with '{'
            elif char == "{":
                self.index += 1
                return self.parse_object()
            # <array> starts with '['
            elif char == "[":
                self.index += 1
                return self.parse_array()
            # there can be an edge case in which a key is empty and at the end of an object
            # like "key": }. We return an empty string here to close the object properly
            elif self.context.current == ContextValues.OBJECT_VALUE and char == "}":
                self.log(
                    "At the end of an object we found a key with missing value, skipping",
                )
                return ""
            # <string> starts with a quote
            elif not self.context.empty and (
                char in self.STRING_DELIMITERS or char.isalpha()
            ):
                return self.parse_string()
            # <number> starts with [0-9] or minus
            elif not self.context.empty and (
                char.isdigit() or char == "-" or char == "."
            ):
                return self.parse_number()
            elif char in ["#", "/"]:
                return self.parse_comment()
            # If everything else fails, we just ignore and move on
            else:
                self.index += 1

    def parse_object(self) -> dict[str, JSONReturnType]:
        # <object> ::= '{' [ <member> *(', ' <member>) ] '}' ; A sequence of 'members'
        obj: dict[str, JSONReturnType] = {}
        # Stop when you either find the closing parentheses or you have iterated over the entire string
        while (self.get_char_at() or "}") != "}":
            # This is what we expect to find:
            # <member> ::= <string> ': ' <json>

            # Skip filler whitespaces
            self.skip_whitespaces_at()

            # Sometimes LLMs do weird things, if we find a ":" so early, we'll change it to "," and move on
            if (self.get_char_at() or "") == ":":
                self.log(
                    "While parsing an object we found a : before a key, ignoring",
                )
                self.index += 1

            # We are now searching for they string key
            # Context is used in the string parser to manage the lack of quotes
            self.context.set(ContextValues.OBJECT_KEY)

            # Save this index in case we need find a duplicate key
            rollback_index = self.index

            # <member> starts with a <string>
            key = ""
            while self.get_char_at():
                # The rollback index needs to be updated here in case the key is empty
                rollback_index = self.index
                if self.get_char_at() == "[" and key == "":
                    # Is this an array?
                    # Need to check if the previous parsed value contained in obj is an array and in that case parse and merge the two
                    prev_key = list(obj.keys())[-1] if obj else None
                    if prev_key and isinstance(obj[prev_key], list):
                        # If the previous key's value is an array, parse the new array and merge
                        self.index += 1
                        new_array = self.parse_array()
                        if isinstance(new_array, list):
                            # Merge and flatten the arrays
                            prev_value = obj[prev_key]
                            if isinstance(prev_value, list):
                                prev_value.extend(
                                    new_array[0]
                                    if len(new_array) == 1
                                    and isinstance(new_array[0], list)
                                    else new_array
                                )
                            self.skip_whitespaces_at()
                            if self.get_char_at() == ",":
                                self.index += 1
                            self.skip_whitespaces_at()
                            continue
                key = str(self.parse_string())
                if key == "":
                    self.skip_whitespaces_at()
                if key != "" or (key == "" and self.get_char_at() in [":", "}"]):
                    # If the string is empty but there is a object divider, we are done here
                    break
            if ContextValues.ARRAY in self.context.context and key in obj:
                self.log(
                    "While parsing an object we found a duplicate key, closing the object here and rolling back the index",
                )
                self.index = rollback_index - 1
                # add an opening curly brace to make this work
                self.json_str = (
                    self.json_str[: self.index + 1]
                    + "{"
                    + self.json_str[self.index + 1 :]
                )
                break

            # Skip filler whitespaces
            self.skip_whitespaces_at()

            # We reached the end here
            if (self.get_char_at() or "}") == "}":
                continue

            self.skip_whitespaces_at()

            # An extreme case of missing ":" after a key
            if (self.get_char_at() or "") != ":":
                self.log(
                    "While parsing an object we missed a : after a key",
                )

            self.index += 1
            self.context.reset()
            self.context.set(ContextValues.OBJECT_VALUE)
            # The value can be any valid json
            value = self.parse_json()

            # Reset context since our job is done
            self.context.reset()
            obj[key] = value

            if (self.get_char_at() or "") in [",", "'", '"']:
                self.index += 1

            # Remove trailing spaces
            self.skip_whitespaces_at()

        self.index += 1
        return obj

    def parse_array(self) -> list[JSONReturnType]:
        # <array> ::= '[' [ <json> *(', ' <json>) ] ']' ; A sequence of JSON values separated by commas
        arr = []
        self.context.set(ContextValues.ARRAY)
        # Stop when you either find the closing parentheses or you have iterated over the entire string
        char = self.get_char_at()
        while char and char not in ["]", "}"]:
            self.skip_whitespaces_at()
            value = self.parse_json()

            # It is possible that parse_json() returns nothing valid, so we increase by 1
            if value == "":
                self.index += 1
            elif value == "..." and self.get_char_at(-1) == ".":
                self.log(
                    "While parsing an array, found a stray '...'; ignoring it",
                )
            else:
                arr.append(value)

            # skip over whitespace after a value but before closing ]
            char = self.get_char_at()
            while char and char != "]" and (char.isspace() or char == ","):
                self.index += 1
                char = self.get_char_at()

        # Especially at the end of an LLM generated json you might miss the last "]"
        if char and char != "]":
            self.log(
                "While parsing an array we missed the closing ], ignoring it",
            )

        self.index += 1

        self.context.reset()
        return arr

    def parse_string(self) -> str | bool | None:
        # <string> is a string of valid characters enclosed in quotes
        # i.e. { name: "John" }
        # Somehow all weird cases in an invalid JSON happen to be resolved in this function, so be careful here

        # Flag to manage corner cases related to missing starting quote
        missing_quotes = False
        doubled_quotes = False
        lstring_delimiter = rstring_delimiter = '"'

        char = self.get_char_at()
        if char in ["#", "/"]:
            return self.parse_comment()
        # A valid string can only start with a valid quote or, in our case, with a literal
        while char and char not in self.STRING_DELIMITERS and not char.isalnum():
            self.index += 1
            char = self.get_char_at()

        if not char:
            # This is an empty string
            return ""

        # Ensuring we use the right delimiter
        if char == "'":
            lstring_delimiter = rstring_delimiter = "'"
        elif char == "“":
            lstring_delimiter = "“"
            rstring_delimiter = "”"
        elif char.isalnum():
            # This could be a <boolean> and not a string. Because (T)rue or (F)alse or (N)ull are valid
            # But remember, object keys are only of type string
            if (
                char.lower() in ["t", "f", "n"]
                and self.context.current != ContextValues.OBJECT_KEY
            ):
                value = self.parse_boolean_or_null()
                if value != "":
                    return value
            self.log(
                "While parsing a string, we found a literal instead of a quote",
            )
            missing_quotes = True

        if not missing_quotes:
            self.index += 1

        # There is sometimes a weird case of doubled quotes, we manage this also later in the while loop
        if (
            self.get_char_at() in self.STRING_DELIMITERS
            and self.get_char_at() == lstring_delimiter
        ):
            # If it's an empty key, this was easy
            if (
                self.context.current == ContextValues.OBJECT_KEY
                and self.get_char_at(1) == ":"
            ):
                self.index += 1
                return ""
            if self.get_char_at(1) == lstring_delimiter:
                # There's something fishy about this, we found doubled quotes and then again quotes
                self.log(
                    "While parsing a string, we found a doubled quote and then a quote again, ignoring it",
                )
                return ""
            # Find the next delimiter
            i = self.skip_to_character(character=rstring_delimiter, idx=1)
            next_c = self.get_char_at(i)
            # Now check that the next character is also a delimiter to ensure that we have "".....""
            # In that case we ignore this rstring delimiter
            if next_c and (self.get_char_at(i + 1) or "") == rstring_delimiter:
                self.log(
                    "While parsing a string, we found a valid starting doubled quote",
                )
                doubled_quotes = True
                self.index += 1
            else:
                # Ok this is not a doubled quote, check if this is an empty string or not
                i = self.skip_whitespaces_at(idx=1, move_main_index=False)
                next_c = self.get_char_at(i)
                if next_c in self.STRING_DELIMITERS + ["{", "["]:
                    # something fishy is going on here
                    self.log(
                        "While parsing a string, we found a doubled quote but also another quote afterwards, ignoring it",
                    )
                    self.index += 1
                    return ""
                elif next_c not in [",", "]", "}"]:
                    self.log(
                        "While parsing a string, we found a doubled quote but it was a mistake, removing one quote",
                    )
                    self.index += 1

        # Initialize our return value
        string_acc = ""

        # Here things get a bit hairy because a string missing the final quote can also be a key or a value in an object
        # In that case we need to use the ":|,|}" characters as terminators of the string
        # So this will stop if:
        # * It finds a closing quote
        # * It iterated over the entire sequence
        # * If we are fixing missing quotes in an object, when it finds the special terminators
        char = self.get_char_at()
        unmatched_delimiter = False
        while char and char != rstring_delimiter:
            if (
                missing_quotes
                and self.context.current == ContextValues.OBJECT_KEY
                and (char == ":" or char.isspace())
            ):
                self.log(
                    "While parsing a string missing the left delimiter in object key context, we found a :, stopping here",
                )
                break
            if (
                (missing_quotes or not self.stream_stable)
                and self.context.current == ContextValues.OBJECT_VALUE
                and char
                in [
                    ",",
                    "}",
                ]
            ):
                rstring_delimiter_missing = True
                # check if this is a case in which the closing comma is NOT missing instead
                i = self.skip_to_character(character=rstring_delimiter, idx=1)
                next_c = self.get_char_at(i)
                if next_c:
                    i += 1
                    # found a delimiter, now we need to check that is followed strictly by a comma or brace
                    # or the string ended
                    i = self.skip_whitespaces_at(idx=i, move_main_index=False)
                    next_c = self.get_char_at(i)
                    if not next_c or next_c in [",", "}"]:
                        rstring_delimiter_missing = False
                    else:
                        # OK but this could still be some garbage at the end of the string
                        # So we need to check if we find a new lstring_delimiter afterwards
                        # If we do, maybe this is a missing delimiter
                        i = self.skip_to_character(character=lstring_delimiter, idx=i)
                        next_c = self.get_char_at(i)
                        if not next_c:
                            rstring_delimiter_missing = False
                        else:
                            # But again, this could just be something a bit stupid like "lorem, "ipsum" sic"
                            # Check if we find a : afterwards (skipping space)
                            i = self.skip_whitespaces_at(
                                idx=i + 1, move_main_index=False
                            )
                            next_c = self.get_char_at(i)
                            if next_c and next_c != ":":
                                rstring_delimiter_missing = False
                else:
                    # There could be a case in which even the next key:value is missing delimeters
                    # because it might be a systemic issue with the output
                    # So let's check if we can find a : in the string instead
                    i = self.skip_to_character(character=":", idx=1)
                    next_c = self.get_char_at(i)
                    if next_c:
                        # OK then this is a systemic issue with the output
                        break
                    else:
                        # skip any whitespace first
                        i = self.skip_whitespaces_at(idx=1, move_main_index=False)
                        # We couldn't find any rstring_delimeter before the end of the string
                        # check if this is the last string of an object and therefore we can keep going
                        # make an exception if this is the last char before the closing brace
                        j = self.skip_to_character(character="}", idx=i)
                        if j - i > 1:
                            # Ok it's not right after the comma
                            # Let's ignore
                            rstring_delimiter_missing = False
                        # Check that j was not out of bound
                        elif self.get_char_at(j):
                            # Check for an unmatched opening brace in string_acc
                            for c in reversed(string_acc):
                                if c == "{":
                                    # Ok then this is part of the string
                                    rstring_delimiter_missing = False
                                    break
                if rstring_delimiter_missing:
                    self.log(
                        "While parsing a string missing the left delimiter in object value context, we found a , or } and we couldn't determine that a right delimiter was present. Stopping here",
                    )
                    break
            if (
                (missing_quotes or not self.stream_stable)
                and char == "]"
                and ContextValues.ARRAY in self.context.context
            ):
                # We found the end of an array and we are in array context
                # So let's check if we find a rstring_delimiter forward otherwise end early
                i = self.skip_to_character(rstring_delimiter)
                if not self.get_char_at(i):
                    # No delimiter found
                    break
            string_acc += char
            self.index += 1
            char = self.get_char_at()
            # Unclosed string ends with a \ character. This character is ignored if stream_stable = True.
            if self.stream_stable and not char and string_acc[-1] == "\\":
                string_acc = string_acc[:-1]
            if char and string_acc[-1] == "\\":
                # This is a special case, if people use real strings this might happen
                self.log("Found a stray escape sequence, normalizing it")
                if char in [rstring_delimiter, "t", "n", "r", "b", "\\"]:
                    string_acc = string_acc[:-1]
                    escape_seqs = {"t": "\t", "n": "\n", "r": "\r", "b": "\b"}
                    string_acc += escape_seqs.get(char, char) or char
                    self.index += 1
                    char = self.get_char_at()
            # If we are in object key context and we find a colon, it could be a missing right quote
            if (
                char == ":"
                and not missing_quotes
                and self.context.current == ContextValues.OBJECT_KEY
            ):
                # Ok now we need to check if this is followed by a value like "..."
                i = self.skip_to_character(character=lstring_delimiter, idx=1)
                next_c = self.get_char_at(i)
                if next_c:
                    i += 1
                    # found the first delimiter
                    i = self.skip_to_character(character=rstring_delimiter, idx=i)
                    next_c = self.get_char_at(i)
                    if next_c:
                        # found a second delimiter
                        i += 1
                        # Skip spaces
                        i = self.skip_whitespaces_at(idx=i, move_main_index=False)
                        next_c = self.get_char_at(i)
                        if next_c and next_c in [",", "}"]:
                            # Ok then this is a missing right quote
                            self.log(
                                "While parsing a string missing the right delimiter in object key context, we found a :, stopping here",
                            )
                            break
                else:
                    # The string ended without finding a lstring_delimiter, I will assume this is a missing right quote
                    self.log(
                        "While parsing a string missing the right delimiter in object key context, we found a :, stopping here",
                    )
                    break
            # ChatGPT sometimes forget to quote stuff in html tags or markdown, so we do this whole thing here
            if char == rstring_delimiter:
                # Special case here, in case of double quotes one after another
                if doubled_quotes and self.get_char_at(1) == rstring_delimiter:
                    self.log(
                        "While parsing a string, we found a doubled quote, ignoring it"
                    )
                    self.index += 1
                elif (
                    missing_quotes
                    and self.context.current == ContextValues.OBJECT_VALUE
                ):
                    # In case of missing starting quote I need to check if the delimeter is the end or the beginning of a key
                    i = 1
                    next_c = self.get_char_at(i)
                    while next_c and next_c not in [
                        rstring_delimiter,
                        lstring_delimiter,
                    ]:
                        i += 1
                        next_c = self.get_char_at(i)
                    if next_c:
                        # We found a quote, now let's make sure there's a ":" following
                        i += 1
                        # found a delimiter, now we need to check that is followed strictly by a comma or brace
                        i = self.skip_whitespaces_at(idx=i, move_main_index=False)
                        next_c = self.get_char_at(i)
                        if next_c and next_c == ":":
                            # Reset the cursor
                            self.index -= 1
                            char = self.get_char_at()
                            self.log(
                                "In a string with missing quotes and object value context, I found a delimeter but it turns out it was the beginning on the next key. Stopping here.",
                            )
                            break
                elif unmatched_delimiter:
                    unmatched_delimiter = False
                    string_acc += str(char)
                    self.index += 1
                    char = self.get_char_at()
                else:
                    # Check if eventually there is a rstring delimiter, otherwise we bail
                    i = 1
                    next_c = self.get_char_at(i)
                    check_comma_in_object_value = True
                    while next_c and next_c not in [
                        rstring_delimiter,
                        lstring_delimiter,
                    ]:
                        # This is a bit of a weird workaround, essentially in object_value context we don't always break on commas
                        # This is because the routine after will make sure to correct any bad guess and this solves a corner case
                        if check_comma_in_object_value and next_c.isalpha():
                            check_comma_in_object_value = False
                        # If we are in an object context, let's check for the right delimiters
                        if (
                            (
                                ContextValues.OBJECT_KEY in self.context.context
                                and next_c in [":", "}"]
                            )
                            or (
                                ContextValues.OBJECT_VALUE in self.context.context
                                and next_c == "}"
                            )
                            or (
                                ContextValues.ARRAY in self.context.context
                                and next_c in ["]", ","]
                            )
                            or (
                                check_comma_in_object_value
                                and self.context.current == ContextValues.OBJECT_VALUE
                                and next_c == ","
                            )
                        ):
                            break
                        i += 1
                        next_c = self.get_char_at(i)
                    # If we stopped for a comma in object_value context, let's check if find a "} at the end of the string
                    if (
                        next_c == ","
                        and self.context.current == ContextValues.OBJECT_VALUE
                    ):
                        i += 1
                        i = self.skip_to_character(character=rstring_delimiter, idx=i)
                        next_c = self.get_char_at(i)
                        # Ok now I found a delimiter, let's skip whitespaces and see if next we find a }
                        i += 1
                        i = self.skip_whitespaces_at(idx=i, move_main_index=False)
                        next_c = self.get_char_at(i)
                    elif (
                        next_c == rstring_delimiter and self.get_char_at(i - 1) != "\\"
                    ):
                        # Check if self.index:self.index+i is only whitespaces, break if that's the case
                        if all(
                            str(self.get_char_at(j)).isspace()
                            for j in range(1, i)
                            if self.get_char_at(j)
                        ):
                            break
                        if self.context.current == ContextValues.OBJECT_VALUE:
                            # But this might not be it! This could be just a missing comma
                            # We found a delimiter and we need to check if this is a key
                            # so find a rstring_delimiter and a colon after
                            i = self.skip_to_character(
                                character=rstring_delimiter, idx=i + 1
                            )
                            i += 1
                            next_c = self.get_char_at(i)
                            while next_c and next_c != ":":
                                if next_c in [",", "]", "}"] or (
                                    next_c == rstring_delimiter
                                    and self.get_char_at(i - 1) != "\\"
                                ):
                                    break
                                i += 1
                                next_c = self.get_char_at(i)
                            # Only if we fail to find a ':' then we know this is misplaced quote
                            if next_c != ":":
                                self.log(
                                    "While parsing a string, we a misplaced quote that would have closed the string but has a different meaning here, ignoring it",
                                )
                                unmatched_delimiter = not unmatched_delimiter
                                string_acc += str(char)
                                self.index += 1
                                char = self.get_char_at()
                        elif self.context.current == ContextValues.ARRAY:
                            # If we got up to here it means that this is a situation like this:
                            # ["bla bla bla "puppy" bla bla bla "kitty" bla bla"]
                            # So we need to ignore this quote
                            self.log(
                                "While parsing a string in Array context, we detected a quoted section that would have closed the string but has a different meaning here, ignoring it",
                            )
                            unmatched_delimiter = not unmatched_delimiter
                            string_acc += str(char)
                            self.index += 1
                            char = self.get_char_at()
                        elif self.context.current == ContextValues.OBJECT_KEY:
                            # In this case we just ignore this and move on
                            self.log(
                                "While parsing a string in Object Key context, we detected a quoted section that would have closed the string but has a different meaning here, ignoring it",
                            )
                            string_acc += str(char)
                            self.index += 1
                            char = self.get_char_at()
        if (
            char
            and missing_quotes
            and self.context.current == ContextValues.OBJECT_KEY
            and char.isspace()
        ):
            self.log(
                "While parsing a string, handling an extreme corner case in which the LLM added a comment instead of valid string, invalidate the string and return an empty value",
            )
            self.skip_whitespaces_at()
            if self.get_char_at() not in [":", ","]:
                return ""

        # A fallout of the previous special case in the while loop,
        # we need to update the index only if we had a closing quote
        if char != rstring_delimiter:
            # if stream_stable = True, unclosed strings do not trim trailing whitespace characters
            if not self.stream_stable:
                self.log(
                    "While parsing a string, we missed the closing quote, ignoring",
                )
                string_acc = string_acc.rstrip()
        else:
            self.index += 1

        if not self.stream_stable and (
            missing_quotes or (string_acc and string_acc[-1] == "\n")
        ):
            # Clean the whitespaces for some corner cases
            string_acc = string_acc.rstrip()

        return string_acc

    def parse_number(self) -> float | int | str | JSONReturnType:
        # <number> is a valid real number expressed in one of a number of given formats
        number_str = ""
        char = self.get_char_at()
        is_array = self.context.current == ContextValues.ARRAY
        while char and char in self.NUMBER_CHARS and (not is_array or char != ","):
            number_str += char
            self.index += 1
            char = self.get_char_at()
        if number_str and number_str[-1] in "-eE/,":
            # The number ends with a non valid character for a number/currency, rolling back one
            number_str = number_str[:-1]
            self.index -= 1
        elif (self.get_char_at() or "").isalpha():
            # this was a string instead, sorry
            self.index -= len(number_str)
            return self.parse_string()
        try:
            if "," in number_str:
                return str(number_str)
            if "." in number_str or "e" in number_str or "E" in number_str:
                return float(number_str)
            else:
                return int(number_str)
        except ValueError:
            return number_str

    def parse_boolean_or_null(self) -> bool | str | None:
        # <boolean> is one of the literal strings 'true', 'false', or 'null' (unquoted)
        starting_index = self.index
        char = (self.get_char_at() or "").lower()
        value: tuple[str, bool | None] | None = None
        if char == "t":
            value = ("true", True)
        elif char == "f":
            value = ("false", False)
        elif char == "n":
            value = ("null", None)

        if value:
            i = 0
            while char and i < len(value[0]) and char == value[0][i]:
                i += 1
                self.index += 1
                char = (self.get_char_at() or "").lower()
            if i == len(value[0]):
                return value[1]

        # If nothing works reset the index before returning
        self.index = starting_index
        return ""

    def parse_comment(self) -> str:
        """
        Parse code-like comments:

        - "# comment": A line comment that continues until a newline.
        - "// comment": A line comment that continues until a newline.
        - "/* comment */": A block comment that continues until the closing delimiter "*/".

        The comment is skipped over and an empty string is returned so that comments do not interfere
        with the actual JSON elements.
        """
        char = self.get_char_at()
        termination_characters = ["\n", "\r"]
        if ContextValues.ARRAY in self.context.context:
            termination_characters.append("]")
        if ContextValues.OBJECT_VALUE in self.context.context:
            termination_characters.append("}")
        if ContextValues.OBJECT_KEY in self.context.context:
            termination_characters.append(":")
        # Line comment starting with #
        if char == "#":
            comment = ""
            while char and char not in termination_characters:
                comment += char
                self.index += 1
                char = self.get_char_at()
            self.log(f"Found line comment: {comment}")
            return ""

        # Comments starting with '/'
        elif char == "/":
            next_char = self.get_char_at(1)
            # Handle line comment starting with //
            if next_char == "/":
                comment = "//"
                self.index += 2  # Skip both slashes.
                char = self.get_char_at()
                while char and char not in termination_characters:
                    comment += char
                    self.index += 1
                    char = self.get_char_at()
                self.log(f"Found line comment: {comment}")
                return ""
            # Handle block comment starting with /*
            elif next_char == "*":
                comment = "/*"
                self.index += 2  # Skip '/*'
                while True:
                    char = self.get_char_at()
                    if not char:
                        self.log(
                            "Reached end-of-string while parsing block comment; unclosed block comment."
                        )
                        break
                    comment += char
                    self.index += 1
                    if comment.endswith("*/"):
                        break
                self.log(f"Found block comment: {comment}")
                return ""
            else:
                # Skip standalone '/' characters that are not part of a comment
                # to avoid getting stuck in an infinite loop
                self.index += 1
                return ""
        return ""  # pragma: no cover

    def get_char_at(self, count: int = 0) -> str | Literal[False]:
        # Why not use something simpler? Because try/except in python is a faster alternative to an "if" statement that is often True
        try:
            return self.json_str[self.index + count]
        except IndexError:
            return False

    def skip_whitespaces_at(self, idx: int = 0, move_main_index=True) -> int:
        """
        This function quickly iterates on whitespaces, syntactic sugar to make the code more concise
        """
        try:
            char = self.json_str[self.index + idx]
        except IndexError:
            return idx
        while char.isspace():
            if move_main_index:
                self.index += 1
            else:
                idx += 1
            try:
                char = self.json_str[self.index + idx]
            except IndexError:
                return idx
        return idx

    def skip_to_character(self, character: str, idx: int = 0) -> int:
        """
        This function quickly iterates to find a character, syntactic sugar to make the code more concise
        """
        try:
            char = self.json_str[self.index + idx]
        except IndexError:
            return idx
        while char != character:
            idx += 1
            try:
                char = self.json_str[self.index + idx]
            except IndexError:
                return idx
        return idx

    def _log(self, text: str) -> None:
        window: int = 10
        start: int = max(self.index - window, 0)
        end: int = min(self.index + window, len(self.json_str))
        context: str = self.json_str[start:end]
        self.logger.append(
            {
                "text": text,
                "context": context,
            }
        )



"""
This module will parse the JSON file following the BNF definition:

    <json> ::= <container>

    <primitive> ::= <number> | <string> | <boolean>
    ; Where:
    ; <number> is a valid real number expressed in one of a number of given formats
    ; <string> is a string of valid characters enclosed in quotes
    ; <boolean> is one of the literal strings 'true', 'false', or 'null' (unquoted)

    <container> ::= <object> | <array>
    <array> ::= '[' [ <json> *(', ' <json>) ] ']' ; A sequence of JSON values separated by commas
    <object> ::= '{' [ <member> *(', ' <member>) ] '}' ; A sequence of 'members'
    <member> ::= <string> ': ' <json> ; A pair consisting of a name, and a JSON value

If something is wrong (a missing parentheses or quotes for example) it will use a few simple heuristics to fix the JSON string:
- Add the missing parentheses if the parser believes that the array or object should be closed
- Quote strings or add missing single quotes
- Adjust whitespaces and remove line breaks

All supported use cases are in the unit tests
"""

import argparse
import json
import sys
from typing import Literal, TextIO, overload

from .json_parser import JSONParser, JSONReturnType


@overload
def repair_json(
    json_str: str = "",
    return_objects: Literal[False] = False,
    skip_json_loads: bool = False,
    logging: bool = False,
    json_fd: TextIO | None = None,
    ensure_ascii: bool = True,
    chunk_length: int = 0,
    stream_stable: bool = False,
) -> str: ...


@overload
def repair_json(
    json_str: str = "",
    return_objects: Literal[True] = True,
    skip_json_loads: bool = False,
    logging: bool = False,
    json_fd: TextIO | None = None,
    ensure_ascii: bool = True,
    chunk_length: int = 0,
    stream_stable: bool = False,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]: ...


def repair_json(
    json_str: str = "",
    return_objects: bool = False,
    skip_json_loads: bool = False,
    logging: bool = False,
    json_fd: TextIO | None = None,
    ensure_ascii: bool = True,
    chunk_length: int = 0,
    stream_stable: bool = False,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
    """
    Given a json formatted string, it will try to decode it and, if it fails, it will try to fix it.

    Args:
        json_str (str, optional): The JSON string to repair. Defaults to an empty string.
        return_objects (bool, optional): If True, return the decoded data structure. Defaults to False.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.
        json_fd (Optional[TextIO], optional): File descriptor for JSON input. Do not use! Use `from_file` or `load` instead. Defaults to None.
        ensure_ascii (bool, optional): Set to False to avoid converting non-latin characters to ascii (for example when using chinese characters). Defaults to True. Ignored if `skip_json_loads` is True.
        chunk_length (int, optional): Size in bytes of the file chunks to read at once. Ignored if `json_fd` is None. Do not use! Use `from_file` or `load` instead. Defaults to 1MB.
        stream_stable (bool, optional): When the json to be repaired is the accumulation of streaming json at a certain moment.If this parameter to True will keep the repair results stable.
    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]]]: The repaired JSON or a tuple with the repaired JSON and repair log.
    """
    parser = JSONParser(json_str, json_fd, logging, chunk_length, stream_stable)
    if skip_json_loads:
        parsed_json = parser.parse()
    else:
        try:
            parsed_json = json.load(json_fd) if json_fd else json.loads(json_str)
        except json.JSONDecodeError:
            parsed_json = parser.parse()
    # It's useful to return the actual object instead of the json string,
    # it allows this lib to be a replacement of the json library
    if return_objects or logging:
        return parsed_json
    # Avoid returning only a pair of quotes if it's an empty string
    elif parsed_json == "":
        return ""
    return json.dumps(parsed_json, ensure_ascii=ensure_ascii)


def loads(
    json_str: str,
    skip_json_loads: bool = False,
    logging: bool = False,
    stream_stable: bool = False,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]] | str:
    """
    This function works like `json.loads()` except that it will fix your JSON in the process.
    It is a wrapper around the `repair_json()` function with `return_objects=True`.

    Args:
        json_str (str): The JSON string to load and repair.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.

    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]], str]: The repaired JSON object or a tuple with the repaired JSON object and repair log.
    """
    return repair_json(
        json_str=json_str,
        return_objects=True,
        skip_json_loads=skip_json_loads,
        logging=logging,
        stream_stable=stream_stable,
    )


def load(
    fd: TextIO,
    skip_json_loads: bool = False,
    logging: bool = False,
    chunk_length: int = 0,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
    """
    This function works like `json.load()` except that it will fix your JSON in the process.
    It is a wrapper around the `repair_json()` function with `json_fd=fd` and `return_objects=True`.

    Args:
        fd (TextIO): File descriptor for JSON input.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.
        chunk_length (int, optional): Size in bytes of the file chunks to read at once. Defaults to 1MB.

    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]]]: The repaired JSON object or a tuple with the repaired JSON object and repair log.
    """
    return repair_json(
        json_fd=fd,
        chunk_length=chunk_length,
        return_objects=True,
        skip_json_loads=skip_json_loads,
        logging=logging,
    )


def from_file(
    filename: str,
    skip_json_loads: bool = False,
    logging: bool = False,
    chunk_length: int = 0,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
    """
    This function is a wrapper around `load()` so you can pass the filename as string

    Args:
        filename (str): The name of the file containing JSON data to load and repair.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.
        chunk_length (int, optional): Size in bytes of the file chunks to read at once. Defaults to 1MB.

    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]]]: The repaired JSON object or a tuple with the repaired JSON object and repair log.
    """
    with open(filename) as fd:
        jsonobj = load(
            fd=fd,
            skip_json_loads=skip_json_loads,
            logging=logging,
            chunk_length=chunk_length,
        )

    return jsonobj


def cli(inline_args: list[str] | None = None) -> int:
    """
    Command-line interface for repairing and parsing JSON files.

    Args:
        inline_args (Optional[List[str]]): List of command-line arguments for testing purposes. Defaults to None.
            - filename (str): The JSON file to repair. If omitted, the JSON is read from stdin.
            - -i, --inline (bool): Replace the file inline instead of returning the output to stdout.
            - -o, --output TARGET (str): If specified, the output will be written to TARGET filename instead of stdout.
            - --ensure_ascii (bool): Pass ensure_ascii=True to json.dumps(). Will pass False otherwise.
            - --indent INDENT (int): Number of spaces for indentation (Default 2).

    Returns:
        int: Exit code of the CLI operation.

    Raises:
        Exception: Any exception that occurs during file processing.

    Example:
        >>> cli(['example.json', '--indent', '4'])
        >>> cat json.txt | json_repair
    """
    parser = argparse.ArgumentParser(description="Repair and parse JSON files.")
    # Make the filename argument optional; if omitted, we will read from stdin.
    parser.add_argument(
        "filename",
        nargs="?",
        help="The JSON file to repair (if omitted, reads from stdin)",
    )
    parser.add_argument(
        "-i",
        "--inline",
        action="store_true",
        help="Replace the file inline instead of returning the output to stdout",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="TARGET",
        help="If specified, the output will be written to TARGET filename instead of stdout",
    )
    parser.add_argument(
        "--ensure_ascii",
        action="store_true",
        help="Pass ensure_ascii=True to json.dumps()",
    )
    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help="Number of spaces for indentation (Default 2)",
    )

    if inline_args is None:  # pragma: no cover
        args = parser.parse_args()
    else:
        args = parser.parse_args(inline_args)

    # Inline mode requires a filename, so error out if none was provided.
    if args.inline and not args.filename:  # pragma: no cover
        print("Error: Inline mode requires a filename", file=sys.stderr)
        sys.exit(1)

    if args.inline and args.output:  # pragma: no cover
        print("Error: You cannot pass both --inline and --output", file=sys.stderr)
        sys.exit(1)

    ensure_ascii = False
    if args.ensure_ascii:
        ensure_ascii = True

    try:
        # Use from_file if a filename is provided; otherwise read from stdin.
        if args.filename:
            result = from_file(args.filename)
        else:
            data = sys.stdin.read()
            result = loads(data)
        if args.inline or args.output:
            with open(args.output or args.filename, mode="w") as fd:
                json.dump(result, fd, indent=args.indent, ensure_ascii=ensure_ascii)
        else:
            print(json.dumps(result, indent=args.indent, ensure_ascii=ensure_ascii))
    except Exception as e:  # pragma: no cover
        print(f"Error: {str(e)}", file=sys.stderr)
        return 1

    return 0  # Success


if __name__ == "__main__":  # pragma: no cover
    sys.exit(cli())



import io
import os.path
import pathlib
import tempfile
from unittest.mock import patch

import pytest

from src.json_repair.json_repair import cli, from_file, loads, repair_json


def test_basic_types_valid():
    assert repair_json("True", return_objects=True) == ""
    assert repair_json("False", return_objects=True) == ""
    assert repair_json("Null", return_objects=True) == ""
    assert repair_json("1", return_objects=True) == 1
    assert repair_json("[]", return_objects=True) == []
    assert repair_json("[1, 2, 3, 4]", return_objects=True) == [1, 2, 3, 4]
    assert repair_json("{}", return_objects=True) == {}
    assert repair_json(
        '{ "key": "value", "key2": 1, "key3": True }', return_objects=True
    ) == {"key": "value", "key2": 1, "key3": True}


def test_basic_types_invalid():
    assert repair_json("true", return_objects=True)
    assert not repair_json("false", return_objects=True)
    assert repair_json("null", return_objects=True) is None
    assert repair_json("1.2", return_objects=True) == 1.2
    assert repair_json("[", return_objects=True) == []
    assert repair_json("[1, 2, 3, 4", return_objects=True) == [1, 2, 3, 4]
    assert repair_json("{", return_objects=True) == {}
    assert repair_json(
        '{ "key": value, "key2": 1 "key3": null }', return_objects=True
    ) == {"key": "value", "key2": 1, "key3": None}


def test_valid_json():
    assert (
        repair_json('{"name": "John", "age": 30, "city": "New York"}')
        == '{"name": "John", "age": 30, "city": "New York"}'
    )
    assert (
        repair_json('{"employees":["John", "Anna", "Peter"]} ')
        == '{"employees": ["John", "Anna", "Peter"]}'
    )
    assert repair_json('{"key": "value:value"}') == '{"key": "value:value"}'
    assert (
        repair_json('{"text": "The quick brown fox,"}')
        == '{"text": "The quick brown fox,"}'
    )
    assert (
        repair_json('{"text": "The quick brown fox won\'t jump"}')
        == '{"text": "The quick brown fox won\'t jump"}'
    )
    assert repair_json('{"key": ""') == '{"key": ""}'
    assert (
        repair_json('{"key1": {"key2": [1, 2, 3]}}') == '{"key1": {"key2": [1, 2, 3]}}'
    )
    assert (
        repair_json('{"key": 12345678901234567890}') == '{"key": 12345678901234567890}'
    )
    assert repair_json('{"key": "value\u263a"}') == '{"key": "value\\u263a"}'
    assert repair_json('{"key": "value\\nvalue"}') == '{"key": "value\\nvalue"}'


def test_brackets_edge_cases():
    assert repair_json("[{]") == "[{}]"
    assert repair_json("   {  }   ") == "{}"
    assert repair_json("[") == "[]"
    assert repair_json("]") == ""
    assert repair_json("{") == "{}"
    assert repair_json("}") == ""
    assert repair_json('{"') == "{}"
    assert repair_json('["') == "[]"
    assert repair_json("{foo: [}") == '{"foo": []}'


def test_general_edge_cases():
    assert repair_json('"') == ""
    assert repair_json("\n") == ""
    assert repair_json(" ") == ""
    assert repair_json("[[1\n\n]") == "[[1]]"
    assert repair_json("string") == ""
    assert repair_json("stringbeforeobject {}") == "{}"
    assert repair_json("/") == ""


def test_mixed_data_types():
    assert (
        repair_json('  {"key": true, "key2": false, "key3": null}')
        == '{"key": true, "key2": false, "key3": null}'
    )
    assert (
        repair_json('{"key": TRUE, "key2": FALSE, "key3": Null}   ')
        == '{"key": true, "key2": false, "key3": null}'
    )


def test_missing_and_mixed_quotes():
    assert (
        repair_json(
            "{'key': 'string', 'key2': false, \"key3\": null, \"key4\": unquoted}"
        )
        == '{"key": "string", "key2": false, "key3": null, "key4": "unquoted"}'
    )
    assert (
        repair_json('{"name": "John", "age": 30, "city": "New York')
        == '{"name": "John", "age": 30, "city": "New York"}'
    )
    assert (
        repair_json('{"name": "John", "age": 30, city: "New York"}')
        == '{"name": "John", "age": 30, "city": "New York"}'
    )
    assert (
        repair_json('{"name": "John", "age": 30, "city": New York}')
        == '{"name": "John", "age": 30, "city": "New York"}'
    )
    assert (
        repair_json('{"name": John, "age": 30, "city": "New York"}')
        == '{"name": "John", "age": 30, "city": "New York"}'
    )
    assert (
        repair_json('{“slanted_delimiter”: "value"}')
        == '{"slanted_delimiter": "value"}'
    )
    assert (
        repair_json('{"name": "John", "age": 30, "city": "New')
        == '{"name": "John", "age": 30, "city": "New"}'
    )
    assert (
        repair_json('{"name": "John", "age": 30, "city": "New York, "gender": "male"}')
        == '{"name": "John", "age": 30, "city": "New York", "gender": "male"}'
    )

    assert (
        repair_json('[{"key": "value", COMMENT "notes": "lorem "ipsum", sic." }]')
        == '[{"key": "value", "notes": "lorem \\"ipsum\\", sic."}]'
    )
    assert repair_json('{"key": ""value"}') == '{"key": "value"}'
    assert (
        repair_json('{"key": "value", 5: "value"}') == '{"key": "value", "5": "value"}'
    )
    assert repair_json('{"foo": "\\"bar\\""') == '{"foo": "\\"bar\\""}'
    assert repair_json('{"" key":"val"') == '{" key": "val"}'
    assert (
        repair_json('{"key": value "key2" : "value2" ')
        == '{"key": "value", "key2": "value2"}'
    )
    assert (
        repair_json('{"key": "lorem ipsum ... "sic " tamet. ...}')
        == '{"key": "lorem ipsum ... \\"sic \\" tamet. ..."}'
    )
    assert repair_json('{"key": value , }') == '{"key": "value"}'
    assert (
        repair_json('{"comment": "lorem, "ipsum" sic "tamet". To improve"}')
        == '{"comment": "lorem, \\"ipsum\\" sic \\"tamet\\". To improve"}'
    )
    assert repair_json('{"key": "v"alu"e"} key:') == '{"key": "v\\"alu\\"e"}'


def test_array_edge_cases():
    assert repair_json("[1, 2, 3,") == "[1, 2, 3]"
    assert repair_json("[1, 2, 3, ...]") == "[1, 2, 3]"
    assert repair_json("[1, 2, ... , 3]") == "[1, 2, 3]"
    assert repair_json("[1, 2, '...', 3]") == '[1, 2, "...", 3]'
    assert repair_json("[true, false, null, ...]") == "[true, false, null]"
    assert repair_json('["a" "b" "c" 1') == '["a", "b", "c", 1]'
    assert (
        repair_json('{"employees":["John", "Anna",')
        == '{"employees": ["John", "Anna"]}'
    )
    assert (
        repair_json('{"employees":["John", "Anna", "Peter')
        == '{"employees": ["John", "Anna", "Peter"]}'
    )
    assert repair_json('{"key1": {"key2": [1, 2, 3') == '{"key1": {"key2": [1, 2, 3]}}'
    assert repair_json('{"key": ["value]}') == '{"key": ["value"]}'
    assert repair_json('["lorem "ipsum" sic"]') == '["lorem \\"ipsum\\" sic"]'
    assert (
        repair_json('{"key1": ["value1", "value2"}, "key2": ["value3", "value4"]}')
        == '{"key1": ["value1", "value2"], "key2": ["value3", "value4"]}'
    )
    assert repair_json('[ "value", /* comment */ "value2" ]') == '["value", "value2"]'
    assert (
        repair_json('{"key": ["value" "value1" "value2"]}')
        == '{"key": ["value", "value1", "value2"]}'
    )
    assert (
        repair_json(
            '{"key": ["lorem "ipsum" dolor "sit" amet, "consectetur" ", "lorem "ipsum" dolor", "lorem"]}'
        )
        == '{"key": ["lorem \\"ipsum\\" dolor \\"sit\\" amet, \\"consectetur\\" ", "lorem \\"ipsum\\" dolor", "lorem"]}'
    )
    assert repair_json('{"k"e"y": "value"}') == '{"k\\"e\\"y": "value"}'


def test_escaping():
    assert repair_json("'\"'") == ""
    assert (
        repair_json('{"key": \'string"\n\t\\le\'') == '{"key": "string\\"\\n\\t\\\\le"}'
    )
    assert (
        repair_json(
            r'{"real_content": "Some string: Some other string \t Some string <a href=\"https://domain.com\">Some link</a>"'
        )
        == r'{"real_content": "Some string: Some other string \t Some string <a href=\"https://domain.com\">Some link</a>"}'
    )
    assert repair_json('{"key_1\n": "value"}') == '{"key_1": "value"}'
    assert repair_json('{"key\t_": "value"}') == '{"key\\t_": "value"}'


def test_object_edge_cases():
    assert repair_json("{       ") == "{}"
    assert repair_json('{"": "value"') == '{"": "value"}'
    assert (
        repair_json('{"value_1": true, COMMENT "value_2": "data"}')
        == '{"value_1": true, "value_2": "data"}'
    )
    assert (
        repair_json('{"value_1": true, SHOULD_NOT_EXIST "value_2": "data" AAAA }')
        == '{"value_1": true, "value_2": "data"}'
    )
    assert (
        repair_json('{"" : true, "key2": "value2"}') == '{"": true, "key2": "value2"}'
    )
    assert (
        repair_json(
            """{""answer"":[{""traits"":''Female aged 60+'',""answer1"":""5""}]}"""
        )
        == '{"answer": [{"traits": "Female aged 60+", "answer1": "5"}]}'
    )
    assert (
        repair_json('{ "words": abcdef", "numbers": 12345", "words2": ghijkl" }')
        == '{"words": "abcdef", "numbers": 12345, "words2": "ghijkl"}'
    )
    assert (
        repair_json("""{"number": 1,"reason": "According...""ans": "YES"}""")
        == '{"number": 1, "reason": "According...", "ans": "YES"}'
    )
    assert repair_json("""{ "a" : "{ b": {} }" }""") == '{"a": "{ b"}'
    assert repair_json("""{"b": "xxxxx" true}""") == '{"b": "xxxxx"}'
    assert (
        repair_json('{"key": "Lorem "ipsum" s,"}') == '{"key": "Lorem \\"ipsum\\" s,"}'
    )
    assert (
        repair_json('{"lorem": ipsum, sic, datum.",}')
        == '{"lorem": "ipsum, sic, datum."}'
    )
    assert (
        repair_json(
            '{"lorem": sic tamet. "ipsum": sic tamet, quick brown fox. "sic": ipsum}'
        )
        == '{"lorem": "sic tamet.", "ipsum": "sic tamet", "sic": "ipsum"}'
    )
    assert (
        repair_json('{"lorem_ipsum": "sic tamet, quick brown fox. }')
        == '{"lorem_ipsum": "sic tamet, quick brown fox."}'
    )
    assert (
        repair_json('{"key":value, " key2":"value2" }')
        == '{"key": "value", " key2": "value2"}'
    )
    assert (
        repair_json('{"key":value "key2":"value2" }')
        == '{"key": "value", "key2": "value2"}'
    )
    assert (
        repair_json("{'text': 'words{words in brackets}more words'}")
        == '{"text": "words{words in brackets}more words"}'
    )
    assert (
        repair_json("{text:words{words in brackets}}")
        == '{"text": "words{words in brackets}"}'
    )
    assert (
        repair_json("{text:words{words in brackets}m}")
        == '{"text": "words{words in brackets}m"}'
    )
    assert repair_json('{"key": "value, value2"```') == '{"key": "value, value2"}'
    assert (
        repair_json("{key:value,key2:value2}") == '{"key": "value", "key2": "value2"}'
    )
    assert repair_json('{"key:"value"}') == '{"key": "value"}'
    assert repair_json('{"key:value}') == '{"key": "value"}'
    assert (
        repair_json('[{"lorem": {"ipsum": "sic"}, """" "lorem": {"ipsum": "sic"}]')
        == '[{"lorem": {"ipsum": "sic"}}, {"lorem": {"ipsum": "sic"}}]'
    )
    assert (
        repair_json('{ "key": { "key2": "value2" // comment }, "key3": "value3" }')
        == '{"key": {"key2": "value2"}, "key3": "value3"}'
    )
    assert (
        repair_json('{ "key": { "key2": "value2" # comment }, "key3": "value3" }')
        == '{"key": {"key2": "value2"}, "key3": "value3"}'
    )
    assert (
        repair_json('{ "key": { "key2": "value2" /* comment */ }, "key3": "value3" }')
        == '{"key": {"key2": "value2"}, "key3": "value3"}'
    )
    assert repair_json('{ "key": "value" /* comment') == '{"key": "value"}'
    assert (
        repair_json(
            '{ "key": ["arrayvalue"], ["arrayvalue1"], ["arrayvalue2"], "key3": "value3" }'
        )
        == '{"key": ["arrayvalue", "arrayvalue1", "arrayvalue2"], "key3": "value3"}'
    )
    assert (
        repair_json('{ "key": ["arrayvalue"], "key3": "value3", ["arrayvalue1"] }')
        == '{"key": ["arrayvalue"], "key3": "value3", "arrayvalue1": ""}'
    )


def test_number_edge_cases():
    assert (
        repair_json(' - { "test_key": ["test_value", "test_value2"] }')
        == '{"test_key": ["test_value", "test_value2"]}'
    )
    assert repair_json('{"key": 1/3}') == '{"key": "1/3"}'
    assert repair_json('{"key": .25}') == '{"key": 0.25}'
    assert (
        repair_json('{"here": "now", "key": 1/3, "foo": "bar"}')
        == '{"here": "now", "key": "1/3", "foo": "bar"}'
    )
    assert repair_json('{"key": 12345/67890}') == '{"key": "12345/67890"}'
    assert repair_json("[105,12") == "[105, 12]"
    assert repair_json('{"key", 105,12,') == '{"key": "105,12"}'
    assert repair_json('{"key": 1/3, "foo": "bar"}') == '{"key": "1/3", "foo": "bar"}'
    assert repair_json('{"key": 10-20}') == '{"key": "10-20"}'
    assert repair_json('{"key": 1.1.1}') == '{"key": "1.1.1"}'
    assert repair_json("[- ") == "[]"
    assert repair_json('{"key": 1. }') == '{"key": 1.0}'
    assert repair_json('{"key": 1e10 }') == '{"key": 10000000000.0}'
    assert repair_json('{"key": 1e }') == '{"key": 1}'
    assert repair_json('{"key": 1notanumber }') == '{"key": "1notanumber"}'
    assert repair_json("[1, 2notanumber]") == '[1, "2notanumber"]'


def test_markdown():
    assert (
        repair_json('{ "content": "[LINK]("https://google.com")" }')
        == '{"content": "[LINK](\\"https://google.com\\")"}'
    )
    assert repair_json('{ "content": "[LINK](" }') == '{"content": "[LINK]("}'
    assert (
        repair_json('{ "content": "[LINK](", "key": true }')
        == '{"content": "[LINK](", "key": true}'
    )


def test_leading_trailing_characters():
    assert repair_json('````{ "key": "value" }```') == '{"key": "value"}'
    assert (
        repair_json("""{    "a": "",    "b": [ { "c": 1} ] \n}```""")
        == '{"a": "", "b": [{"c": 1}]}'
    )
    assert (
        repair_json(
            "Based on the information extracted, here is the filled JSON output: ```json { 'a': 'b' } ```"
        )
        == '{"a": "b"}'
    )
    assert (
        repair_json("""
                       The next 64 elements are:
                       ```json
                       { "key": "value" }
                       ```""")
        == '{"key": "value"}'
    )


def test_multiple_jsons():
    assert repair_json("[]{}") == "[[], {}]"
    assert repair_json("{}[]{}") == "[{}, [], {}]"
    assert (
        repair_json('{"key":"value"}[1,2,3,True]')
        == '[{"key": "value"}, [1, 2, 3, true]]'
    )
    assert (
        repair_json(
            'lorem ```json {"key":"value"} ``` ipsum ```json [1,2,3,True] ``` 42'
        )
        == '[{"key": "value"}, [1, 2, 3, true]]'
    )
    assert (
        repair_json('[{"key":"value"}][{"key":"value_after"}]')
        == '[{"key": "value_after"}]'
    )


def test_repair_json_with_objects():
    # Test with valid JSON strings
    assert repair_json("[]", return_objects=True) == []
    assert repair_json("{}", return_objects=True) == {}
    assert repair_json(
        '{"key": true, "key2": false, "key3": null}', return_objects=True
    ) == {"key": True, "key2": False, "key3": None}
    assert repair_json(
        '{"name": "John", "age": 30, "city": "New York"}', return_objects=True
    ) == {
        "name": "John",
        "age": 30,
        "city": "New York",
    }
    assert repair_json("[1, 2, 3, 4]", return_objects=True) == [1, 2, 3, 4]
    assert repair_json(
        '{"employees":["John", "Anna", "Peter"]} ', return_objects=True
    ) == {"employees": ["John", "Anna", "Peter"]}
    assert repair_json(
        """
{
  "resourceType": "Bundle",
  "id": "1",
  "type": "collection",
  "entry": [
    {
      "resource": {
        "resourceType": "Patient",
        "id": "1",
        "name": [
          {"use": "official", "family": "Corwin", "given": ["Keisha", "Sunny"], "prefix": ["Mrs."},
          {"use": "maiden", "family": "Goodwin", "given": ["Keisha", "Sunny"], "prefix": ["Mrs."]}
        ]
      }
    }
  ]
}
""",
        return_objects=True,
    ) == {
        "resourceType": "Bundle",
        "id": "1",
        "type": "collection",
        "entry": [
            {
                "resource": {
                    "resourceType": "Patient",
                    "id": "1",
                    "name": [
                        {
                            "use": "official",
                            "family": "Corwin",
                            "given": ["Keisha", "Sunny"],
                            "prefix": ["Mrs."],
                        },
                        {
                            "use": "maiden",
                            "family": "Goodwin",
                            "given": ["Keisha", "Sunny"],
                            "prefix": ["Mrs."],
                        },
                    ],
                }
            }
        ],
    }
    assert repair_json(
        '{\n"html": "<h3 id="aaa">Waarom meer dan 200 Technical Experts - "Passie voor techniek"?</h3>"}',
        return_objects=True,
    ) == {
        "html": '<h3 id="aaa">Waarom meer dan 200 Technical Experts - "Passie voor techniek"?</h3>'
    }
    assert repair_json(
        """
        [
            {
                "foo": "Foo bar baz",
                "tag": "#foo-bar-baz"
            },
            {
                "foo": "foo bar "foobar" foo bar baz.",
                "tag": "#foo-bar-foobar"
            }
        ]
        """,
        return_objects=True,
    ) == [
        {"foo": "Foo bar baz", "tag": "#foo-bar-baz"},
        {"foo": 'foo bar "foobar" foo bar baz.', "tag": "#foo-bar-foobar"},
    ]


def test_repair_json_skip_json_loads():
    assert (
        repair_json('{"key": true, "key2": false, "key3": null}', skip_json_loads=True)
        == '{"key": true, "key2": false, "key3": null}'
    )
    assert repair_json(
        '{"key": true, "key2": false, "key3": null}',
        return_objects=True,
        skip_json_loads=True,
    ) == {"key": True, "key2": False, "key3": None}
    assert (
        repair_json('{"key": true, "key2": false, "key3": }', skip_json_loads=True)
        == '{"key": true, "key2": false, "key3": ""}'
    )
    assert loads('{"key": true, "key2": false, "key3": }', skip_json_loads=True) == {
        "key": True,
        "key2": False,
        "key3": "",
    }


def test_repair_json_from_file():
    path = pathlib.Path(__file__).parent.resolve()

    # Use chunk_length 2 to test the buffering feature
    assert from_file(filename=os.path.join(path, "invalid.json")) == [
        {
            "_id": "655b66256574f09bdae8abe8",
            "index": 0,
            "guid": "31082ae3-b0f3-4406-90f4-cc450bd4379d",
            "isActive": False,
            "balance": "$2,562.78",
            "picture": "http://placehold.it/32x32",
            "age": 32,
            "eyeColor": "brown",
            "name": "Glover Rivas",
            "gender": "male",
            "company": "EMPIRICA",
            "email": "gloverrivas@empirica.com",
            "phone": "+1 (842) 507-3063",
            "address": "536 Montague Terrace, Jenkinsville, Kentucky, 2235",
            "about": "Mollit consectetur excepteur voluptate tempor dolore ullamco enim irure ullamco non enim officia. Voluptate occaecat proident laboris ea Lorem cupidatat reprehenderit nisi nisi aliqua. Amet nulla ipsum deserunt excepteur amet ad aute aute ex. Et enim minim sit veniam est quis dolor nisi sunt quis eiusmod in. Amet eiusmod cillum sunt occaecat dolor laboris voluptate in eiusmod irure aliqua duis.",
            "registered": "2023-11-18T09:32:36 -01:00",
            "latitude": 36.26102,
            "longitude": -91.304608,
            "tags": ["non", "tempor", "do", "ullamco", "dolore", "sunt", "ipsum"],
            "friends": [
                {"id": 0, "name": "Cara Shepherd"},
                {"id": 1, "name": "Mason Farley"},
                {"id": 2, "name": "Harriet Cochran"},
            ],
            "greeting": "Hello, Glover Rivas! You have 7 unread messages.",
            "favoriteFruit": "strawberry",
        },
        {
            "_id": "655b662585364bc57278bb6f",
            "index": 1,
            "guid": "0dea7a3a-f812-4dde-b78d-7a9b58e5da05",
            "isActive": True,
            "balance": "$1,359.48",
            "picture": "http://placehold.it/32x32",
            "age": 38,
            "eyeColor": "brown",
            "name": "Brandi Moreno",
            "gender": "female",
            "company": "MARQET",
            "email": "brandimoreno@marqet.com",
            "phone": "+1 (850) 434-2077",
            "address": "537 Doone Court, Waiohinu, Michigan, 3215",
            "about": "Irure proident adipisicing do Lorem do incididunt in laborum in eiusmod eiusmod ad elit proident. Eiusmod dolor ex magna magna occaecat. Nulla deserunt velit ex exercitation et irure sunt. Cupidatat ut excepteur ea quis labore sint cupidatat incididunt amet eu consectetur cillum ipsum proident. Occaecat exercitation aute laborum dolor proident reprehenderit laborum in voluptate culpa. Exercitation nulla adipisicing culpa aute est deserunt ea nisi deserunt consequat occaecat ut et non. Incididunt ex exercitation dolor dolor anim cillum dolore.",
            "registered": "2015-09-03T11:47:15 -02:00",
            "latitude": -19.768953,
            "longitude": 8.948458,
            "tags": [
                "laboris",
                "occaecat",
                "laborum",
                "laborum",
                "ex",
                "cillum",
                "occaecat",
            ],
            "friends": [
                {"id": 0, "name": "Erna Kelly"},
                {"id": 1, "name": "Black Mays"},
                {"id": 2, "name": "Davis Buck"},
            ],
            "greeting": "Hello, Brandi Moreno! You have 1 unread messages.",
            "favoriteFruit": "apple",
        },
        {
            "_id": "655b6625870da431bcf5e0c2",
            "index": 2,
            "guid": "b17f6e3f-c898-4334-abbf-05cf222f143b",
            "isActive": False,
            "balance": "$1,493.77",
            "picture": "http://placehold.it/32x32",
            "age": 20,
            "eyeColor": "brown",
            "name": "Moody Meadows",
            "gender": "male",
            "company": "OPTIQUE",
            "email": "moodymeadows@optique.com",
            "phone": "+1 (993) 566-3041",
            "address": "766 Osborn Street, Bath, Maine, 7666",
            "about": "Non commodo excepteur nostrud qui adipisicing aliquip dolor minim nulla culpa proident. In ad cupidatat ea mollit ex est do deserunt proident nostrud. Cillum id id eiusmod amet exercitation nostrud cillum sunt deserunt dolore deserunt eiusmod mollit. Ut ex tempor ad laboris voluptate labore id officia fugiat exercitation amet.",
            "registered": "2015-01-16T02:48:28 -01:00",
            "latitude": -25.847327,
            "longitude": 63.95991,
            "tags": [
                "aute",
                "commodo",
                "adipisicing",
                "nostrud",
                "duis",
                "mollit",
                "ut",
            ],
            "friends": [
                {"id": 0, "name": "Lacey Cash"},
                {"id": 1, "name": "Gabrielle Harmon"},
                {"id": 2, "name": "Ellis Lambert"},
            ],
            "greeting": "Hello, Moody Meadows! You have 4 unread messages.",
            "favoriteFruit": "strawberry",
        },
        {
            "_id": "655b6625f3e1bf422220854e",
            "index": 3,
            "guid": "92229883-2bfd-4974-a08c-1b506b372e46",
            "isActive": False,
            "balance": "$2,215.34",
            "picture": "http://placehold.it/32x32",
            "age": 22,
            "eyeColor": "brown",
            "name": "Heath Nguyen",
            "gender": "male",
            "company": "BLEENDOT",
            "email": "heathnguyen@bleendot.com",
            "phone": "+1 (989) 512-2797",
            "address": "135 Milton Street, Graniteville, Nebraska, 276",
            "about": "Consequat aliquip irure Lorem cupidatat nulla magna ullamco nulla voluptate adipisicing anim consectetur tempor aliquip. Magna aliqua nulla eu tempor esse proident. Proident fugiat ad ex Lorem reprehenderit dolor aliquip labore labore aliquip. Deserunt aute enim ea minim officia anim culpa sint commodo. Cillum consectetur excepteur aliqua exercitation Lorem veniam voluptate.",
            "registered": "2016-07-06T01:31:07 -02:00",
            "latitude": -60.997048,
            "longitude": -102.397885,
            "tags": ["do", "ad", "consequat", "irure", "tempor", "elit", "minim"],
            "friends": [
                {"id": 0, "name": "Walker Hernandez"},
                {"id": 1, "name": "Maria Lane"},
                {"id": 2, "name": "Mcknight Barron"},
            ],
            "greeting": "Hello, Heath Nguyen! You have 4 unread messages.",
            "favoriteFruit": "apple",
        },
        {
            "_id": "655b6625519a5b5e4b6742bf",
            "index": 4,
            "guid": "c5dc685f-6d0d-4173-b4cf-f5df29a1e8ef",
            "isActive": True,
            "balance": "$1,358.90",
            "picture": "http://placehold.it/32x32",
            "age": 33,
            "eyeColor": "brown",
            "name": "Deidre Duke",
            "gender": "female",
            "company": "OATFARM",
            "email": "deidreduke@oatfarm.com",
            "phone": "+1 (875) 587-3256",
            "address": "487 Schaefer Street, Wattsville, West Virginia, 4506",
            "about": "Laboris eu nulla esse magna sit eu deserunt non est aliqua exercitation commodo. Ad occaecat qui qui laborum dolore anim Lorem. Est qui occaecat irure enim deserunt enim aliqua ex deserunt incididunt esse. Quis in minim laboris proident non mollit. Magna ea do labore commodo. Et elit esse esse occaecat officia ipsum nisi.",
            "registered": "2021-09-12T04:17:08 -02:00",
            "latitude": 68.609781,
            "longitude": -87.509134,
            "tags": [
                "mollit",
                "cupidatat",
                "irure",
                "sit",
                "consequat",
                "anim",
                "fugiat",
            ],
            "friends": [
                {"id": 0, "name": "Bean Paul"},
                {"id": 1, "name": "Cochran Hubbard"},
                {"id": 2, "name": "Rodgers Atkinson"},
            ],
            "greeting": "Hello, Deidre Duke! You have 6 unread messages.",
            "favoriteFruit": "apple",
        },
        {
            "_id": "655b6625a19b3f7e5f82f0ea",
            "index": 5,
            "guid": "75f3c264-baa1-47a0-b21c-4edac23d9935",
            "isActive": True,
            "balance": "$3,554.36",
            "picture": "http://placehold.it/32x32",
            "age": 26,
            "eyeColor": "blue",
            "name": "Lydia Holland",
            "gender": "female",
            "company": "ESCENTA",
            "email": "lydiaholland@escenta.com",
            "phone": "+1 (927) 482-3436",
            "address": "554 Rockaway Parkway, Kohatk, Montana, 6316",
            "about": "Consectetur ea est labore commodo laborum mollit pariatur non enim. Est dolore et non laboris tempor. Ea incididunt ut adipisicing cillum labore officia tempor eiusmod commodo. Cillum fugiat ex consectetur ut nostrud anim nostrud exercitation ut duis in ea. Eu et id fugiat est duis eiusmod ullamco quis officia minim sint ea nisi in.",
            "registered": "2018-03-13T01:48:56 -01:00",
            "latitude": -88.495799,
            "longitude": 71.840667,
            "tags": [
                "veniam",
                "minim",
                "consequat",
                "consequat",
                "incididunt",
                "consequat",
                "elit",
            ],
            "friends": [
                {"id": 0, "name": "Debra Massey"},
                {"id": 1, "name": "Weiss Savage"},
                {"id": 2, "name": "Shannon Guerra"},
            ],
            "greeting": "Hello, Lydia Holland! You have 5 unread messages.",
            "favoriteFruit": "banana",
        },
    ]
    assert from_file(filename=os.path.join(path, "invalid.json"), chunk_length=2) == [
        {
            "_id": "655b66256574f09bdae8abe8",
            "index": 0,
            "guid": "31082ae3-b0f3-4406-90f4-cc450bd4379d",
            "isActive": False,
            "balance": "$2,562.78",
            "picture": "http://placehold.it/32x32",
            "age": 32,
            "eyeColor": "brown",
            "name": "Glover Rivas",
            "gender": "male",
            "company": "EMPIRICA",
            "email": "gloverrivas@empirica.com",
            "phone": "+1 (842) 507-3063",
            "address": "536 Montague Terrace, Jenkinsville, Kentucky, 2235",
            "about": "Mollit consectetur excepteur voluptate tempor dolore ullamco enim irure ullamco non enim officia. Voluptate occaecat proident laboris ea Lorem cupidatat reprehenderit nisi nisi aliqua. Amet nulla ipsum deserunt excepteur amet ad aute aute ex. Et enim minim sit veniam est quis dolor nisi sunt quis eiusmod in. Amet eiusmod cillum sunt occaecat dolor laboris voluptate in eiusmod irure aliqua duis.",
            "registered": "2023-11-18T09:32:36 -01:00",
            "latitude": 36.26102,
            "longitude": -91.304608,
            "tags": ["non", "tempor", "do", "ullamco", "dolore", "sunt", "ipsum"],
            "friends": [
                {"id": 0, "name": "Cara Shepherd"},
                {"id": 1, "name": "Mason Farley"},
                {"id": 2, "name": "Harriet Cochran"},
            ],
            "greeting": "Hello, Glover Rivas! You have 7 unread messages.",
            "favoriteFruit": "strawberry",
        },
        {
            "_id": "655b662585364bc57278bb6f",
            "index": 1,
            "guid": "0dea7a3a-f812-4dde-b78d-7a9b58e5da05",
            "isActive": True,
            "balance": "$1,359.48",
            "picture": "http://placehold.it/32x32",
            "age": 38,
            "eyeColor": "brown",
            "name": "Brandi Moreno",
            "gender": "female",
            "company": "MARQET",
            "email": "brandimoreno@marqet.com",
            "phone": "+1 (850) 434-2077",
            "address": "537 Doone Court, Waiohinu, Michigan, 3215",
            "about": "Irure proident adipisicing do Lorem do incididunt in laborum in eiusmod eiusmod ad elit proident. Eiusmod dolor ex magna magna occaecat. Nulla deserunt velit ex exercitation et irure sunt. Cupidatat ut excepteur ea quis labore sint cupidatat incididunt amet eu consectetur cillum ipsum proident. Occaecat exercitation aute laborum dolor proident reprehenderit laborum in voluptate culpa. Exercitation nulla adipisicing culpa aute est deserunt ea nisi deserunt consequat occaecat ut et non. Incididunt ex exercitation dolor dolor anim cillum dolore.",
            "registered": "2015-09-03T11:47:15 -02:00",
            "latitude": -19.768953,
            "longitude": 8.948458,
            "tags": [
                "laboris",
                "occaecat",
                "laborum",
                "laborum",
                "ex",
                "cillum",
                "occaecat",
            ],
            "friends": [
                {"id": 0, "name": "Erna Kelly"},
                {"id": 1, "name": "Black Mays"},
                {"id": 2, "name": "Davis Buck"},
            ],
            "greeting": "Hello, Brandi Moreno! You have 1 unread messages.",
            "favoriteFruit": "apple",
        },
        {
            "_id": "655b6625870da431bcf5e0c2",
            "index": 2,
            "guid": "b17f6e3f-c898-4334-abbf-05cf222f143b",
            "isActive": False,
            "balance": "$1,493.77",
            "picture": "http://placehold.it/32x32",
            "age": 20,
            "eyeColor": "brown",
            "name": "Moody Meadows",
            "gender": "male",
            "company": "OPTIQUE",
            "email": "moodymeadows@optique.com",
            "phone": "+1 (993) 566-3041",
            "address": "766 Osborn Street, Bath, Maine, 7666",
            "about": "Non commodo excepteur nostrud qui adipisicing aliquip dolor minim nulla culpa proident. In ad cupidatat ea mollit ex est do deserunt proident nostrud. Cillum id id eiusmod amet exercitation nostrud cillum sunt deserunt dolore deserunt eiusmod mollit. Ut ex tempor ad laboris voluptate labore id officia fugiat exercitation amet.",
            "registered": "2015-01-16T02:48:28 -01:00",
            "latitude": -25.847327,
            "longitude": 63.95991,
            "tags": [
                "aute",
                "commodo",
                "adipisicing",
                "nostrud",
                "duis",
                "mollit",
                "ut",
            ],
            "friends": [
                {"id": 0, "name": "Lacey Cash"},
                {"id": 1, "name": "Gabrielle Harmon"},
                {"id": 2, "name": "Ellis Lambert"},
            ],
            "greeting": "Hello, Moody Meadows! You have 4 unread messages.",
            "favoriteFruit": "strawberry",
        },
        {
            "_id": "655b6625f3e1bf422220854e",
            "index": 3,
            "guid": "92229883-2bfd-4974-a08c-1b506b372e46",
            "isActive": False,
            "balance": "$2,215.34",
            "picture": "http://placehold.it/32x32",
            "age": 22,
            "eyeColor": "brown",
            "name": "Heath Nguyen",
            "gender": "male",
            "company": "BLEENDOT",
            "email": "heathnguyen@bleendot.com",
            "phone": "+1 (989) 512-2797",
            "address": "135 Milton Street, Graniteville, Nebraska, 276",
            "about": "Consequat aliquip irure Lorem cupidatat nulla magna ullamco nulla voluptate adipisicing anim consectetur tempor aliquip. Magna aliqua nulla eu tempor esse proident. Proident fugiat ad ex Lorem reprehenderit dolor aliquip labore labore aliquip. Deserunt aute enim ea minim officia anim culpa sint commodo. Cillum consectetur excepteur aliqua exercitation Lorem veniam voluptate.",
            "registered": "2016-07-06T01:31:07 -02:00",
            "latitude": -60.997048,
            "longitude": -102.397885,
            "tags": ["do", "ad", "consequat", "irure", "tempor", "elit", "minim"],
            "friends": [
                {"id": 0, "name": "Walker Hernandez"},
                {"id": 1, "name": "Maria Lane"},
                {"id": 2, "name": "Mcknight Barron"},
            ],
            "greeting": "Hello, Heath Nguyen! You have 4 unread messages.",
            "favoriteFruit": "apple",
        },
        {
            "_id": "655b6625519a5b5e4b6742bf",
            "index": 4,
            "guid": "c5dc685f-6d0d-4173-b4cf-f5df29a1e8ef",
            "isActive": True,
            "balance": "$1,358.90",
            "picture": "http://placehold.it/32x32",
            "age": 33,
            "eyeColor": "brown",
            "name": "Deidre Duke",
            "gender": "female",
            "company": "OATFARM",
            "email": "deidreduke@oatfarm.com",
            "phone": "+1 (875) 587-3256",
            "address": "487 Schaefer Street, Wattsville, West Virginia, 4506",
            "about": "Laboris eu nulla esse magna sit eu deserunt non est aliqua exercitation commodo. Ad occaecat qui qui laborum dolore anim Lorem. Est qui occaecat irure enim deserunt enim aliqua ex deserunt incididunt esse. Quis in minim laboris proident non mollit. Magna ea do labore commodo. Et elit esse esse occaecat officia ipsum nisi.",
            "registered": "2021-09-12T04:17:08 -02:00",
            "latitude": 68.609781,
            "longitude": -87.509134,
            "tags": [
                "mollit",
                "cupidatat",
                "irure",
                "sit",
                "consequat",
                "anim",
                "fugiat",
            ],
            "friends": [
                {"id": 0, "name": "Bean Paul"},
                {"id": 1, "name": "Cochran Hubbard"},
                {"id": 2, "name": "Rodgers Atkinson"},
            ],
            "greeting": "Hello, Deidre Duke! You have 6 unread messages.",
            "favoriteFruit": "apple",
        },
        {
            "_id": "655b6625a19b3f7e5f82f0ea",
            "index": 5,
            "guid": "75f3c264-baa1-47a0-b21c-4edac23d9935",
            "isActive": True,
            "balance": "$3,554.36",
            "picture": "http://placehold.it/32x32",
            "age": 26,
            "eyeColor": "blue",
            "name": "Lydia Holland",
            "gender": "female",
            "company": "ESCENTA",
            "email": "lydiaholland@escenta.com",
            "phone": "+1 (927) 482-3436",
            "address": "554 Rockaway Parkway, Kohatk, Montana, 6316",
            "about": "Consectetur ea est labore commodo laborum mollit pariatur non enim. Est dolore et non laboris tempor. Ea incididunt ut adipisicing cillum labore officia tempor eiusmod commodo. Cillum fugiat ex consectetur ut nostrud anim nostrud exercitation ut duis in ea. Eu et id fugiat est duis eiusmod ullamco quis officia minim sint ea nisi in.",
            "registered": "2018-03-13T01:48:56 -01:00",
            "latitude": -88.495799,
            "longitude": 71.840667,
            "tags": [
                "veniam",
                "minim",
                "consequat",
                "consequat",
                "incididunt",
                "consequat",
                "elit",
            ],
            "friends": [
                {"id": 0, "name": "Debra Massey"},
                {"id": 1, "name": "Weiss Savage"},
                {"id": 2, "name": "Shannon Guerra"},
            ],
            "greeting": "Hello, Lydia Holland! You have 5 unread messages.",
            "favoriteFruit": "banana",
        },
    ]

    # Create a temporary file
    temp_fd, temp_path = tempfile.mkstemp(suffix=".json")
    try:
        # Write content to the temporary file
        with os.fdopen(temp_fd, "w") as tmp:
            tmp.write("{key:value}")
        assert from_file(filename=temp_path, logging=True) == (
            {"key": "value"},
            [
                {
                    "text": "While parsing a string, we found a literal instead of a quote",
                    "context": "{key:value}",
                },
                {
                    "context": "{key:value}",
                    "text": "While parsing a string missing the left delimiter in object key context, we found a :, stopping here",
                },
                {
                    "text": "While parsing a string, we missed the closing quote, ignoring",
                    "context": "{key:value}",
                },
                {
                    "text": "While parsing a string, we found a literal instead of a quote",
                    "context": "{key:value}",
                },
                {
                    "context": "{key:value}",
                    "text": "While parsing a string missing the left delimiter in object value context, we found a , or } and we couldn't determine that a right delimiter was present. Stopping here",
                },
                {
                    "text": "While parsing a string, we missed the closing quote, ignoring",
                    "context": "{key:value}",
                },
            ],
        )
        assert from_file(filename=temp_path, logging=True, chunk_length=2) == (
            {"key": "value"},
            [
                {
                    "text": "While parsing a string, we found a literal instead of a quote",
                    "context": "{key:value}",
                },
                {
                    "context": "{key:value}",
                    "text": "While parsing a string missing the left delimiter in object key context, we found a :, stopping here",
                },
                {
                    "text": "While parsing a string, we missed the closing quote, ignoring",
                    "context": "{key:value}",
                },
                {
                    "text": "While parsing a string, we found a literal instead of a quote",
                    "context": "{key:value}",
                },
                {
                    "context": "{key:value}",
                    "text": "While parsing a string missing the left delimiter in object value context, we found a , or } and we couldn't determine that a right delimiter was present. Stopping here",
                },
                {
                    "text": "While parsing a string, we missed the closing quote, ignoring",
                    "context": "{key:value}",
                },
            ],
        )
    finally:
        # Clean up - delete the temporary file
        os.remove(temp_path)

    # Create a temporary file
    temp_fd, temp_path = tempfile.mkstemp(suffix=".json")
    try:
        # Write content to the temporary file
        with os.fdopen(temp_fd, "w") as tmp:
            tmp.write("x" * 5 * 1024 * 1024)  # 5 MB
        assert from_file(filename=temp_path, logging=True) == ("", [])

    finally:
        # Clean up - delete the temporary file
        os.remove(temp_path)


def test_ensure_ascii():
    assert (
        repair_json("{'test_中国人_ascii':'统一码'}", ensure_ascii=False)
        == '{"test_中国人_ascii": "统一码"}'
    )


def test_stream_stable():
    # default: stream_stable = False
    # When the json to be repaired is the accumulation of streaming json at a certain moment.
    # The default repair result is unstable.
    assert repair_json('{"key": "val\\', stream_stable=False) == '{"key": "val\\\\"}'
    assert repair_json('{"key": "val\\n', stream_stable=False) == '{"key": "val"}'
    assert (
        repair_json('{"key": "val\\n123,`key2:value2', stream_stable=False)
        == '{"key": "val\\n123", "key2": "value2"}'
    )
    assert (
        repair_json('{"key": "val\\n123,`key2:value2`"}', stream_stable=True)
        == '{"key": "val\\n123,`key2:value2`"}'
    )
    # stream_stable = True
    assert repair_json('{"key": "val\\', stream_stable=True) == '{"key": "val"}'
    assert repair_json('{"key": "val\\n', stream_stable=True) == '{"key": "val\\n"}'
    assert (
        repair_json('{"key": "val\\n123,`key2:value2', stream_stable=True)
        == '{"key": "val\\n123,`key2:value2"}'
    )
    assert (
        repair_json('{"key": "val\\n123,`key2:value2`"}', stream_stable=True)
        == '{"key": "val\\n123,`key2:value2`"}'
    )


def test_cli(capsys):
    # Create a temporary file
    temp_fd, temp_path = tempfile.mkstemp(suffix=".json")
    try:
        # Write content to the temporary file
        with os.fdopen(temp_fd, "w") as tmp:
            tmp.write("{key:value")
        cli(inline_args=[temp_path, "--indent", 0, "--ensure_ascii"])
        captured = capsys.readouterr()
        assert captured.out == '{\n"key": "value"\n}\n'

        # Test the output option
        tempout_fd, tempout_path = tempfile.mkstemp(suffix=".json")
        cli(inline_args=[temp_path, "--indent", 0, "-o", tempout_path])
        with open(tempout_path) as tmp:
            out = tmp.read()
        assert out == '{\n"key": "value"\n}'

        # Test the inline option
        cli(inline_args=[temp_path, "--indent", 0, "-i"])
        with open(temp_path) as tmp:
            out = tmp.read()
        assert out == '{\n"key": "value"\n}'

    finally:
        # Clean up - delete the temporary file
        os.remove(temp_path)
        os.remove(tempout_path)

    # Prepare a JSON string that needs to be repaired.
    test_input = "{key:value"
    # Expected output when running cli with --indent 0.
    expected_output = '{\n"key": "value"\n}\n'
    # Patch sys.stdin so that cli() reads from it instead of a file.
    with patch("sys.stdin", new=io.StringIO(test_input)):
        cli(inline_args=["--indent", 0])
    captured = capsys.readouterr()
    assert captured.out == expected_output


def test_cli_inline_requires_filename(capsys):
    """cli() should exit with an error when --inline is passed without a filename."""
    with pytest.raises(SystemExit) as exc:
        cli(inline_args=["--inline"])
    captured = capsys.readouterr()
    assert captured.err.strip() == "Error: Inline mode requires a filename"
    assert exc.value.code != 0


def test_cli_inline_and_output_error(tmp_path, capsys):
    """cli() should exit with an error when --inline and --output are used together."""
    outfile = tmp_path / "out.json"
    with pytest.raises(SystemExit) as exc:
        cli(inline_args=["dummy.json", "--inline", "--output", str(outfile)])
    captured = capsys.readouterr()
    assert captured.err.strip() == "Error: You cannot pass both --inline and --output"
    assert exc.value.code != 0



from typing import Any


class ObjectComparer:  # pragma: no cover
    def __init__(self) -> None:
        pass  # No operation performed in the constructor

    @staticmethod
    def is_same_object(obj1: Any, obj2: Any, path: str = "") -> bool:
        """
        Recursively compares two objects and ensures that:
        - Their types match
        - Their keys/structure match
        """
        if type(obj1) is not type(obj2):
            # Fail immediately if the types don't match
            return False

        if isinstance(obj1, dict):
            # Quick length check before key compares
            if len(obj1) != len(obj2):
                return False
            for key in obj1:
                if key not in obj2:
                    return False
                if not ObjectComparer.is_same_object(obj1[key], obj2[key]):
                    return False
            return True

        elif isinstance(obj1, list):
            if len(obj1) != len(obj2):
                return False
            for i in range(len(obj1)):
                if not ObjectComparer.is_same_object(obj1[i], obj2[i]):
                    return False
            return True

        # For atoms: types already match, so just return True
        return True
